#
# story-point-acc.R,  7 Jan 23
#
# Analysis for blog post: Evaluating Story point estimation error

library("actuar")
library("colorspace")
library("extraDistr")
library("plyr")


par(bty="l")
par(las=1)
par(pch="+")
pal_col=rainbow(2)


# Geometric mean
geo_mean=function(vals) exp(mean(log(vals+1e-5)))-1e-5

# Base scoring functions
SE=function(x, y) (x-y)^2
AE=function(x, y) abs(x-y)
gen_rel_err=function(x, y, beta) abs(1-(y/x)^beta)
APE=function(x, y) gen_rel_err(x, y, -1)
RE=function(x, y) gen_rel_err(x, y, 1)

# Bespoke scoring function
APE_m7=function(x, y) gen_rel_err(x, y, -0.7)

# Evaluate the scoring functions
# Geometric mean is appropriate for ratios, but there is the issue
# of exact estimates having zero error.
metric_vals=function(x, y) return(data.frame(se=mean(SE(x, y)), ae=mean(AE(x, y)),
						ape=geo_mean(APE(x, y)), re=geo_mean(RE(x, y))))

# Ideal estimator has spread of actuals rounds to closest value.
# Calculate the estimation accuracy for the estimate SP,
# based on the number of tasks having this estimate,
# number of tasks modelled by act_mod
ideal_SP_accuracy=function(SP_seq, SP_spread=(-30:30)/60, act_mod)
{
sp_accuracy=function(SP)
   {
   # Model is hour based, so minutes have to be in fractions of an hour
   est_hr=SP+SP_spread
   # The expected number of tasks at each estimation point
   num_tasks=round(predict(act_mod, newdata=data.frame(hours=est_hr)))
   total_tasks=sum(num_tasks)

   # For each possibility, create a given number of tasks having a given actual 'time'
   # rep(est_hr, times=num_tasks)

   return(data.frame(ideal=rep(est_hr, times=num_tasks), est=rep(SP, total_tasks)))
   }

seq_est=adply(SP_seq, .margin=1, sp_accuracy)
seq_metrics=metric_vals(seq_est$ideal, seq_est$est)
return(seq_metrics)
}


# The less ideal case, with the distribution of estimates being
# a percentage of the SP value, and number of tasks having a
# triangle distribution
triangle_SP_accuracy=function(SP_seq, act_mod)
{
sp_accuracy=function(SP)
   {
   # Model is hour based, so minutes have to be in fractions of an hour
   est_hr=SP+seq(-SP*0.5, SP*0.5, by = 1/60)
   t=predict(act_mod, newdata=data.frame(hours=SP))
   # Approximate a triangle distribution
   tbp=c(t/10, t, t/10)
   hours=c(SP*0.5, SP, SP*1.5)
   tri_mod=glm(tbp ~ poly(hours, 3, raw=TRUE))

   # The expected number of tasks at each estimation point
   num_tasks=round(predict(tri_mod, newdata=data.frame(hours=est_hr)))
   total_tasks=sum(num_tasks)

   # For each possibility, create a given number of tasks having a given actual 'time'
   # rep(est_hr, times=num_tasks)

   return(data.frame(ideal=rep(est_hr, times=num_tasks), est=rep(SP, total_tasks)))
   }

seq_est=adply(SP_seq, .margin=1, sp_accuracy)
seq_metrics=metric_vals(seq_est$ideal, seq_est$est)
return(seq_metrics)
}


# Plot range of triangle distributions
triangle_SP_points=function(SP_seq, act_mod)
{
sp_accuracy=function(SP)
   {
   # Model is hour based, so minutes have to be in fractions of an hour
   est_hr=SP+seq(-SP*0.5, SP*0.5, by = 1/60)
   t=predict(act_mod, newdata=data.frame(hours=SP))
   # Approximate a triangle distribution
   tbp=c(t/10, t, t/10)
   hours=c(SP*0.5, SP, SP*1.5)
   tri_mod=glm(tbp ~ poly(hours, 3, raw=TRUE))

   # The expected number of tasks at each estimation point
   num_tasks=round(predict(tri_mod, newdata=data.frame(hours=est_hr)))
   points(est_hr, num_tasks, col=pal_col[SP], pch=".", cex=3)
   }

seq_est=adply(SP_seq, .margin=1, sp_accuracy)
}



hours=1:14
lin_SP=1:5
fib_SP=c(1, 2, 3, 5, 8, 13)

# Zero estimates cannot occur, so use zero-truncated distributions
act_geom=1000*dztgeom(hours, prob=0.4)
act_nbinom=1000*dztnbinom(hours, size=2.2, prob=0.48)
plot(hours, act_geom, log="x", col=pal_col[2], type="b",
	xlab="Actual (some units)", ylab="Tasks")
points(hours, act_nbinom, col=pal_col[1], type="b",)
legend("bottomleft", legend=c("zero-truncated Negative binomial", "zero-truncated Geometric"),
			fill=pal_col, border="white", bty="n")

geom_mod=glm(act_geom ~ poly(hours, 5, raw=TRUE))
nbinom_mod=glm(act_nbinom ~ poly(hours, 5, raw=TRUE))

# Check that fits are good enough
lines(hours, predict(geom_mod))
lines(hours, predict(nbinom_mod))

# Plot distributions, with rounding range in distinct colors, and minute dots.
pal_col=rainbow(5)
plot(hours, act_geom, log="x", type="n",
	xlim=c(0.5, 10), ylim=c(0, 500),
	xlab="Actual (some units)", ylab="Tasks")
a_ply(lin_SP, .margin=1,
		function(SP, act_mod)
		{
   		est_hr=SP+(-30:30/60)
   				# The expected number of tasks at each estimation point
   		num_tasks=round(predict(act_mod, newdata=data.frame(hours=est_hr)))
		points(est_hr, num_tasks, col=pal_col[SP], pch=".", cex=3)
		},
		geom_mod)
text(1.5, 400, "Geometric")
a_ply(lin_SP, .margin=1,
		function(SP, act_mod)
		{
   		est_hr=SP+(-30:30/60)
   				# The expected number of tasks at each estimation point
   		num_tasks=round(predict(act_mod, newdata=data.frame(hours=est_hr)))
		points(est_hr, num_tasks, col=pal_col[SP], pch=".", cex=3)
		lines(est_hr, num_tasks, col=pal_col[SP])
		},
		nbinom_mod)
text(1.0, 240, "Negative Binomial")

# Plot distribution of triangle points with minute dots
plot(hours, act_geom, log="x", type="n",
	xlim=c(0.5, 10), ylim=c(0, 400),
	xlab="Actual (some units)", ylab="Tasks")
triangle_SP_points(lin_SP, geom_mod)
triangle_SP_points(lin_SP, nbinom_mod)



ideal_SP_accuracy(lin_SP, geom_mod)
#           se        ae        ape        re
# 1 0.08684983 0.2554667 0.09798485 0.1034166
ideal_SP_accuracy(lin_SP, nbinom_mod)
#           se        ae        ape         re
# 1 0.08611908 0.2541106 0.08381301 0.08643975

triangle_SP_accuracy(lin_SP, geom_mod)
#          se        ae      ape        re
# 0.5214218 0.5470728 0.130708 0.1345921
triangle_SP_accuracy(lin_SP, nbinom_mod)
#          se        ae       ape       re
# 0.6190887 0.6089933 0.1318798 0.135797


